@@ -1257,6 +1257,8 @@ namespace libtorrent
 		}
 		else if (url_seeds && url_seeds->type() == lazy_entry::list_t)
 		{
+			// only add a URL once
+			std::set<std::string> unique;
 			for (int i = 0, end(url_seeds->list_size()); i < end; ++i)
 			{
 				lazy_entry const* url = url_seeds->list_at(i);
@@ -1265,6 +1267,8 @@ namespace libtorrent
 				web_seed_entry ent(maybe_url_encode(url->string_value())
 					, web_seed_entry::url_seed);
 				if (m_multifile && ent.url[ent.url.size()-1] != '/') ent.url += '/';
+				if (unique.count(ent.url)) continue;
+				unique.insert(ent.url);
 				m_web_seeds.push_back(ent);
 			}
 		}
@@ -1278,12 +1282,16 @@ namespace libtorrent
 		}
 		else if (http_seeds && http_seeds->type() == lazy_entry::list_t)
 		{
+			// only add a URL once
+			std::set<std::string> unique;
 			for (int i = 0, end(http_seeds->list_size()); i < end; ++i)
 			{
 				lazy_entry const* url = http_seeds->list_at(i);
 				if (url->type() != lazy_entry::string_t || url->string_length() == 0) continue;
-				m_web_seeds.push_back(web_seed_entry(maybe_url_encode(url->string_value())
-					, web_seed_entry::http_seed));
+				std::string u = maybe_url_encode(url->string_value());
+				if (unique.count(u)) continue;
+				unique.insert(u);
+				m_web_seeds.push_back(web_seed_entry(u, web_seed_entry::http_seed));
 			}
 		}
 
